from ..initializers import RandomInitializer, NearestNeighbor


class SimulatedAnnealing:
    def __init__(self, graph, t_max, t_min, max_time, init="random", plot=False):
        assert t_max > t_min > 0, "t_max > t_min > 0"

        self.graph = graph
        self.t_max = t_max
        self.t_min = t_min
        self.t = self.t_max
        self.max_time = max_time
        self.init = init
        self.initializer = None
        self.plot = plot
        if self.init == "random":
            self.initializer = RandomInitializer(self.graph, 1)
        elif self.init == "nearest":
            self.initializer = NearestNeighbor(self.graph, 1)

        self.solution = self.initializer.get_init()[0]
        self.cost = self.graph.get_cost(self.solution)



from src.methods import (
    AntColonyOptimization,
    GeneticAlgorithm,
    Graph,
    LinearIntegerProgram,
    ParticleSwarmOptimization,
    TwoOpt,
)

sizes = [5, 10, 25, 50, 75, 100, 150, 200, 250, 500, 750, 1000]
times = [3600, 3600, 3600, 3600, 3600, 3600, 3600, 3600, 3600, 3600, 3600, 3600]
iterations = [1000, 1000, 1000, 1000, 1000, 1000, 2000, 2000, 2000, 3000, 3000, 3000]
population = [10, 10, 10, 10, 10, 10, 25, 25, 25, 50, 50, 50]

# sizes = [15]
# times = [3600]
# iterations = [1000]
# population = [10]

for i in range(len(sizes)):
    size = sizes[i]
    max_time = times[i]
    num_iterations = iterations[i]
    population_size = population[i]

    graph = Graph(size)
    graph.random_complete_graph()
    # graph.plot()
    # graph.show()

    pso = ParticleSwarmOptimization(
        graph,
        iterations=num_iterations,
        population_size=population_size,
        alpha=0.7,
        beta=0.7,
        max_time=max_time,
        init="nearest",
    )
    pso.run()

    acs = AntColonyOptimization(
        graph=graph,
        iterations=num_iterations,
        population_size=population_size,
        max_time=max_time,
    )
    acs.run()

    ga = GeneticAlgorithm(
        graph=graph,
        max_generations=num_iterations,
        population_size=population_size,
        mutation_probability=0.1,
        init="nearest",
        max_time=max_time,
    )
    ga.run()

    lp = LinearIntegerProgram(graph=graph, max_time=max_time)
    lp.build_model()
    lp.run()

    two_opt = TwoOpt(graph, max_time=max_time, init="random")
    two_opt.run()

    print("PSO time: ", pso.get_time())
    print("ACO time: ", acs.get_time())
    print("GA time: ", ga.get_time())
    print("LP time: ", lp.get_time())
    print("Two opt time: ", two_opt.get_time())
    pso.get_best()
    acs.get_best()
    ga.get_best()
    lp.show()
    two_opt.get_best()

    file = "test.txt"
    text = "Graph: " + str(size) + " nodes\n"
    with open(file, "a") as myfile:
        myfile.write("-----------------------------\n")
        myfile.write(text)
        myfile.close()

    pso.save(file)
    acs.save(file)
    ga.save(file)
    lp.save(file)
    two_opt.save(file)



import random
import copy


class NearestNeighbor:
    def __init__(self, graph, population):
        self.graph = graph
        self.population = population

    def get_init(self):
        random_paths = []
        list_vertices = list(self.graph.vertices)

        for _ in range(self.population):
            random_start = random.sample(list_vertices, 1)[0]
            aux_list = copy.copy(list_vertices)
            aux_list.remove(random_start)
            random_list = [random_start]
            while len(aux_list) > 0:
                min_dist = float("Inf")
                next_node = None
                for i in aux_list:
                    if self.graph.edges[(random_list[-1], i)].cost < min_dist:
                        min_dist = self.graph.edges[(random_list[-1], i)].cost
                        next_node = i

                random_list.append(next_node)
                aux_list.remove(next_node)

            random_paths.append(random_list)

        return random_paths



import copy
import datetime

from ..initializers import NearestNeighbor, RandomInitializer


class TwoOpt:
    def __init__(self, graph, max_time, init="random", plot=False):
        self.graph = graph
        self.max_time = max_time
        self.init = init
        self.plot = plot
        self.initializer = None
        self.cost = float("Inf")
        self.count = 0
        self.time = None
        if self.init == "random":
            self.initializer = RandomInitializer(self.graph, 1)
        elif self.init == "nearest":
            self.initializer = NearestNeighbor(self.graph, 1)

        self.solution = self.initializer.get_init()[0]
        self.cost = self.graph.get_cost(self.solution)

    def get_best(self):
        print("\nTwo opt:")
        print(
            "Best solution: ",
            str(self.solution),
            "\t|\tCost: ",
            str(round(self.cost, 2)),
            "\t|\t Swaps: ",
            str(self.count),
        )

    def run(self):
        initial_time = datetime.datetime.utcnow()
        while (datetime.datetime.utcnow() - initial_time).seconds <= self.max_time:
            swap = False
            start = 0
            end = len(self.solution) + 1

            for i in range(len(self.solution) - 1):
                for j in range(i + 2, len(self.solution) + 1):
                    if j - i > len(self.solution) - 2:
                        continue

                    if i == start:
                        part_1 = self.solution[start:j]
                        part_2 = self.solution[j:end]
                        part_2.reverse()
                        aux_solution = part_1 + part_2

                    elif j == end - 1:
                        part_1 = self.solution[start:i]
                        part_2 = self.solution[i:end]
                        part_2.reverse()
                        aux_solution = part_1 + part_2

                    else:
                        part_1 = self.solution[start:i]
                        part_2 = self.solution[i:j]
                        part_2.reverse()
                        part_3 = self.solution[j:end]
                        aux_solution = part_1 + part_2 + part_3

                    aux_cost = self.graph.get_cost(aux_solution)

                    if self.cost > aux_cost:
                        self.solution = aux_solution
                        self.cost = aux_cost
                        swap = True
                        self.count += 1
                        break

                if swap:
                    break

            if not swap:
                break

        self.time = datetime.datetime.utcnow() - initial_time

    def swap_nodes(self):
        initial_time = datetime.datetime.utcnow()
        while (datetime.datetime.utcnow() - initial_time).seconds <= self.max_time:
            swap = False
            for i in range(len(self.solution)):
                for j in range(len(self.solution)):
                    if i == j:
                        continue

                    aux_solution = copy.copy(self.solution)
                    aux_solution[i], aux_solution[j] = aux_solution[j], aux_solution[i]
                    aux_cost = self.graph.get_cost(aux_solution)

                    if self.cost > aux_cost:
                        self.solution = aux_solution
                        self.cost = aux_cost
                        swap = True
                        self.count += 1
                        break

                if swap:
                    break

            if not swap:
                break

    def save(self, file):
        text = (
            "Two-Opt. Best solution: "
            + str(self.solution)
            + "\t|\tCost: "
            + str(self.cost)
            + "\t|\tSwaps: "
            + str(self.count)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)

    def get_time(self):
        return self.time



import copy
import datetime
import random

from ..initializers import NearestNeighbor, RandomInitializer


class Particle:
    def __init__(self, solution, cost):
        self.solution = solution
        self.cost = cost
        self.best_solution = solution
        self.best_cost = cost

        self.velocity = []

    def reset_velocity(self):
        del self.velocity[:]


class ParticleSwarmOptimization:
    def __init__(
        self,
        graph,
        iterations,
        population_size,
        alpha=1,
        beta=1,
        max_time=60,
        init="random",
        plot=False,
    ):
        self.graph = graph
        self.iterations = iterations
        self.population_size = population_size
        self.alpha = alpha
        self.beta = beta
        self.particles = list()
        self.best = None
        self.max_time = max_time
        self.time = None
        self.plot = plot
        self.init = init
        self.initializer = None

        if self.init == "random":
            self.initializer = RandomInitializer(self.graph, self.population_size)
        elif self.init == "nearest":
            self.initializer = NearestNeighbor(self.graph, self.population_size)

        solutions = self.initializer.get_init()

        for solution in solutions:
            particle = Particle(solution=solution, cost=graph.get_cost(solution))
            self.particles.append(particle)

        self.size_population = len(self.particles)

    def get_best(self):
        best = self.particles[0]
        for particle in self.particles:
            if particle.cost < best.cost:
                best = particle
        print("\nPSO:")
        print("Best solution: ", str(best.solution), "\t|\tcost: ", str(best.cost))

    def show(self):
        print("\nSOLUTION:")
        print(
            "Particles: "
            + str(self.size_population)
            + ". Iterations: "
            + str(self.iterations)
        )
        for particle in self.particles:
            print(
                "Best solution: %s\t|\tcost: %d"
                % (str(particle.best_solution), particle.best_cost)
            )

    def evaluate(self):
        costs = [i.cost for i in self.particles]
        mean_cost = sum(costs) / len(costs)
        min_cost = min(costs)

        if min_cost * 1.01 > mean_cost:
            print("BREAK")
            return True
        else:
            return False

    def run(self):
        start = datetime.datetime.utcnow()
        frac = 0.1
        plot_interval = [10 * x for x in range(1, self.iterations // 10 + 1)]
        initial_time = datetime.datetime.utcnow()

        for i in range(self.iterations):
            if i / self.iterations >= frac:
                print("Iteration ", str(i), " of ", str(self.iterations))
                frac += 0.1

            if i >= self.iterations // 5:
                if self.evaluate():
                    break

            self.best = min(self.particles, key=lambda p: p.best_cost)

            for particle in self.particles:
                particle.reset_velocity()
                temp_velocity = []
                best_population = copy.copy(self.best.best_solution)
                best_particle = particle.best_solution
                solution_particle = particle.solution

                for v in range(self.graph.number_vertices):
                    if solution_particle[v] != best_particle[v]:
                        swap = (
                            v,
                            best_particle.index(solution_particle[v]),
                            self.alpha,
                        )

                        temp_velocity.append(swap)

                        aux = best_particle[swap[0]]
                        best_particle[swap[0]] = best_particle[swap[1]]
                        best_particle[swap[1]] = aux

                for v in range(self.graph.number_vertices):
                    if solution_particle[v] != best_population[v]:
                        swap = (
                            v,
                            best_population.index(solution_particle[v]),
                            self.beta,
                        )

                        temp_velocity.append(swap)

                        aux = best_population[swap[0]]
                        best_population[swap[0]] = best_population[swap[1]]
                        best_population[swap[1]] = aux

                particle.velocity = temp_velocity

                for w in temp_velocity:
                    if random.random() <= w[2]:
                        aux = solution_particle[w[0]]
                        solution_particle[w[0]] = solution_particle[w[1]]
                        solution_particle[w[1]] = aux

                particle.solution = solution_particle

                current_cost = self.graph.get_cost(solution_particle)
                particle.cost = current_cost

                if current_cost < particle.best_cost:
                    particle.best_cost = current_cost
                    particle.best_solution = solution_particle

            if self.plot:
                if (i + 1) in plot_interval:
                    filename = (
                        "plots/pso_"
                        + str(self.graph.number_vertices)
                        + "_"
                        + str(i + 1)
                        + ".png"
                    )
                    title = (
                        "Particle Swarm Optimization. Iteration "
                        + str(i + 1)
                        + "\n Solution cost: "
                        + str(round(particle.best_cost, 2))
                    )
                    self.graph.plot_solution(
                        particle.best_solution, filename=filename, title=title
                    )

            if (datetime.datetime.utcnow() - initial_time).seconds > self.max_time:
                break

        self.time = datetime.datetime.utcnow() - start

    def save(self, file):
        best = self.particles[0]
        for particle in self.particles:
            if particle.cost < best.cost:
                best = particle
        text = (
            "PSO. Best solution: "
            + str(best.solution)
            + "\t|\tCost: "
            + str(best.cost)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)

    def get_time(self):
        return self.time



import datetime

from pulp import *


class LinearIntegerProgram:
    def __init__(self, graph, max_time, plot=False):
        self.graph = graph
        self.model = LpProblem("TSP problem", LpMinimize)
        self.v01Travels = LpVariable.dicts(
            "Travel", ((edge[0], edge[1]) for edge in self.graph.edges), cat="Binary"
        )
        self.vOneTour = LpVariable.dicts(
            "Aux", (v for v in self.graph.vertices), cat="Integer"
        )
        self.vertice_list = list(self.graph.vertices)
        self.temp_solution = list()
        self.solution = list()
        self.cost = 0
        self.max_time = max_time
        self.time = None
        self.plot = plot

    def build_model(self):

        for v in self.graph.vertices:
            self.model += (
                lpSum(self.v01Travels[v, x] for x in self.graph.vertices if x != v) == 1
            )
            self.model += (
                lpSum(self.v01Travels[x, v] for x in self.graph.vertices if x != v) == 1
            )

        for i in self.vertice_list:
            for j in self.vertice_list:
                if i != j and i != self.vertice_list[0] and j != self.vertice_list[0]:
                    self.model += (
                        self.vOneTour[i]
                        - self.vOneTour[j]
                        + self.graph.number_vertices * self.v01Travels[i, j]
                        <= self.graph.number_vertices - 1
                    )

        for i in self.vertice_list:
            if i != self.vertice_list[0]:
                self.model += self.vOneTour[i] <= self.graph.number_vertices - 1

        self.model += lpSum(
            self.v01Travels[edge[0], edge[1]] * self.graph.edges[edge].cost
            for edge in self.graph.edges
        )

    def run(self):
        start = datetime.datetime.utcnow()
        # self.model.solve(
        #     GUROBI_CMD(
        #         msg=0,
        #         options=[
        #             ("TimeLimit", self.max_time),
        #             ("MIPGap", 0.05),
        #             ("MIPGapAbs", 0.05),
        #         ],
        #     )
        # )
        self.model.solve(PULP_CBC_CMD(msg=1, fracGap=0.05, maxSeconds=self.max_time))

        print(self.model.status)
        if self.model.status == 1:
            self.solution = list()
            self.temp_solution = list()
            for v in self.model.variables():
                if v.varValue > 0 and "Travel" in v.name:
                    aux = v.name[7:]
                    aux = aux.replace("_", "", 1)
                    self.temp_solution.append(aux)

            array = self.temp_solution[0].split(",")
            self.solution.append(int(array[0][1:]))
            self.solution.append(int(array[1][:-1]))

            del self.temp_solution[0]
            while True:
                for i in range(len(self.temp_solution)):
                    array = self.temp_solution[i].split(",")
                    if array[0][1:] == str(self.solution[-1]):
                        self.solution.append(int(array[1][:-1]))
                        del self.temp_solution[i]
                        break

                if len(self.temp_solution) == 0:
                    break

            del self.solution[-1]

            self.cost = self.graph.get_cost(self.solution)

            if self.plot:
                filename = "plots/lp_" + str(self.graph.number_vertices) + ".png"
                title = (
                    "Mathematical optimization "
                    + "\n Solution cost: "
                    + str(round(self.cost, 2))
                )
                self.graph.plot_solution(
                    self.solution, pheromones=False, filename=filename, title=title
                )

        else:
            self.cost = float("-inf")

        self.time = datetime.datetime.utcnow() - start

    def show(self):
        print("\nLinear programming formulation:")
        print("Best solution: " + str(self.solution) + "\t|\tcost: " + str(self.cost))

    def save(self, file):
        text = (
            "LP. Best solution: "
            + str(self.solution)
            + "\t|\tCost: "
            + str(self.cost)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)

    def get_time(self):
        return self.time



import random
from math import pow, sqrt

import matplotlib.pyplot as plt

plt.ioff()


class Edge:
    def __init__(self, origin, destination, cost, initial_pheromone=0):
        self.origin = origin
        self.destination = destination
        self.cost = cost
        self.pheromone = initial_pheromone


class Graph:
    def __init__(self, vertices):
        self.edges = dict()
        self.vertices = set()
        self.vertices_coordinates = dict()
        self.number_vertices = vertices
        self.x_coordinates = list()
        self.y_coordinates = list()

    def edge_exists(self, origin, destination):
        if (origin, destination) in self.edges.keys():
            return True
        else:
            return False

    def add_edge(self, origin, destination, cost=0.0):
        if not self.edge_exists(origin, destination):
            self.edges[(origin, destination)] = Edge(origin, destination, cost, 0)

    def calculate_distance(self, origin, destination):
        distance = sqrt(
            pow(
                self.vertices_coordinates[origin][0]
                - self.vertices_coordinates[destination][0],
                2,
            )
            + pow(
                self.vertices_coordinates[origin][1]
                - self.vertices_coordinates[destination][1],
                2,
            )
        )
        return distance

    def show(self):
        print("GRAPH:\n")
        for edge in self.edges:
            print(
                "%d linked with %d with cost %d"
                % (edge[0], edge[1], self.edges[edge].cost)
            )

    def plot(self):
        plt.close()
        plt.scatter(self.x_coordinates, self.y_coordinates, c="#000000")
        plt.show()

    def plot_solution(self, path, pheromones=False, filename="none.png", title=""):

        plt.close()

        if pheromones:
            min_pheromone = 100
            max_pheromone = 0
            for i in self.edges:
                if min_pheromone > self.edges[i].pheromone:
                    min_pheromone = self.edges[i].pheromone

                if max_pheromone < self.edges[i].pheromone:
                    max_pheromone = self.edges[i].pheromone

            pheromone_dict = {
                key: 0.0
                + (self.edges[key].pheromone - min_pheromone)
                * 0.4
                / (max_pheromone - min_pheromone)
                for key in self.edges
            }

            for i in pheromone_dict:
                x_values = [
                    self.vertices_coordinates[i[0]][0],
                    self.vertices_coordinates[i[1]][0],
                ]
                y_values = [
                    self.vertices_coordinates[i[0]][1],
                    self.vertices_coordinates[i[1]][1],
                ]
                plt.plot(
                    x_values,
                    y_values,
                    c="#FF0000",
                    alpha=pheromone_dict[i],
                    linewidth=3,
                )

        plt.scatter(self.x_coordinates, self.y_coordinates, c="#000000")
        for i in range(self.number_vertices - 1):
            x_values = [
                self.vertices_coordinates[path[i]][0],
                self.vertices_coordinates[path[i + 1]][0],
            ]
            y_values = [
                self.vertices_coordinates[path[i]][1],
                self.vertices_coordinates[path[i + 1]][1],
            ]
            plt.plot(x_values, y_values, c="#000000")

        x_values = [
            self.vertices_coordinates[path[self.number_vertices - 1]][0],
            self.vertices_coordinates[path[0]][0],
        ]
        y_values = [
            self.vertices_coordinates[path[self.number_vertices - 1]][1],
            self.vertices_coordinates[path[0]][1],
        ]
        plt.plot(x_values, y_values, c="#000000")

        plt.title(title)
        plt.savefig(filename)

    def random_complete_graph(self):

        for i in range(self.number_vertices):
            self.vertices.add(i)
            new = True
            while new:
                x = random.randint(0, self.number_vertices * 2)
                y = random.randint(0, self.number_vertices * 2)
                repeat = False

                for j in self.vertices_coordinates:
                    if (
                        self.vertices_coordinates[j][0] == x
                        and self.vertices_coordinates[j][1] == y
                    ):
                        repeat = True
                        break

                if not repeat:
                    new = False

            self.vertices_coordinates[i] = [x, y]

        for i in self.vertices:
            for j in self.vertices:
                if i < j:
                    cost = self.calculate_distance(i, j)
                    self.add_edge(i, j, cost)
                    self.add_edge(j, i, cost)

        self.x_coordinates = [
            self.vertices_coordinates[i][0] for i in self.vertices_coordinates
        ]
        self.y_coordinates = [
            self.vertices_coordinates[i][1] for i in self.vertices_coordinates
        ]

    def get_random_path(self, size):
        random_paths = []
        list_vertices = list(self.vertices)

        for i in range(size):
            random_list = random.sample(list_vertices, len(list_vertices))
            random_paths.append(random_list)

        return random_paths

    def get_cost(self, path):
        total_cost = 0
        for i in range(self.number_vertices - 1):
            total_cost += self.edges[(path[i], path[i + 1])].cost

        total_cost += self.edges[(path[self.number_vertices - 1], path[0])].cost
        return total_cost

    def update_pheromone(self, edge, pheromone):
        self.edges[edge].pheromone = pheromone



import random


class RandomInitializer:
    def __init__(self, graph, population):
        self.graph = graph
        self.population = population

    def get_init(self):
        random_paths = []
        list_vertices = list(self.graph.vertices)

        for _ in range(self.population):
            random_list = random.sample(list_vertices, len(list_vertices))
            random_paths.append(random_list)

        return random_paths



import copy
import datetime
import random

from ..initializers import NearestNeighbor, RandomInitializer


class Individual:
    """
    The genes are the solution (path)
    """

    def __init__(self, genes, cost, idx):
        self.genes = genes
        self.cost = cost
        self.number_vertices = len(genes)
        self.id = idx

    def random_sub_solution(self):
        start = random.randrange(0, self.number_vertices)
        end = start
        while (
            end == start
            or abs(end - start) == 1
            or abs(end - start) == len(self.genes) - 1
        ):
            end = random.randrange(0, self.number_vertices)
        if start > end:
            start, end = end, start
        return self.genes[start:end]

    def remove_vertice(self, x):
        if x in self.genes:
            del self.genes[self.genes.index(x)]

    def insert_sub_solution(self, sub_solution, x):
        if x != -1:
            self.genes = self.genes[: x + 1] + sub_solution + self.genes[x + 1 :]
        else:
            self.genes = self.genes + sub_solution

    def mutate(self):
        i = random.randrange(0, self.number_vertices)
        j = random.randrange(0, self.number_vertices)
        while i == j:
            j = random.randrange(0, self.number_vertices)

        aux = self.genes[i]
        self.genes[i] = self.genes[j]
        self.genes[j] = aux

    def __str__(self):
        return str(self.genes)


class GeneticAlgorithm:
    def __init__(
        self,
        graph,
        max_generations,
        population_size,
        mutation_probability,
        max_time,
        init="random",
        plot=False,
    ):
        """ """
        self.individuals = list()
        self.graph = graph
        self.max_generations = max_generations
        self.population_size = population_size
        self.mutation_probability = mutation_probability
        self.mating_pool = []
        self.children = []
        self.id = 1
        self.max_time = max_time
        self.time = None

        self.best_genes = None
        self.best_cost = float("inf")
        self.plot = plot
        self.init = init
        self.initializer = None
        if self.init == "random":
            self.initializer = RandomInitializer(self.graph, self.population_size)
        elif self.init == "nearest":
            self.initializer = NearestNeighbor(self.graph, self.population_size)

        solutions = self.initializer.get_init()

        for solution in solutions:
            individual = Individual(
                genes=solution, cost=self.graph.get_cost(solution), idx=self.id
            )
            self.id += 1
            self.individuals.append(individual)

    def get_best(self):
        best = self.individuals[0]
        for individual in self.individuals:
            if individual.cost < best.cost:
                best = individual
        print("\nGA:")
        print("Best solution: ", str(best.genes), "\t|\tcost: ", str(best.cost))

    def update_best(self):
        for individual in self.individuals:
            if individual.cost < self.best_cost:
                self.best_genes = individual.genes
                self.best_cost = individual.cost

    def best_insertion(self, child, sub_solution):
        start = sub_solution[0]
        end = sub_solution[1]
        best_payoff = float("-inf")
        j = 0
        for i in range(0, len(child.genes) - 1):
            payoff = (
                self.graph.edges[(child.genes[i], child.genes[i + 1])].cost
                - self.graph.edges[(child.genes[i], start)].cost
                - self.graph.edges[(end, child.genes[i + 1])].cost
            )

            if payoff > best_payoff:
                best_payoff = payoff
                j = i

        payoff = (
            self.graph.edges[(child.genes[-1], child.genes[0])].cost
            - self.graph.edges[(child.genes[-1], start)].cost
            - self.graph.edges[(end, child.genes[0])].cost
        )

        if payoff > best_payoff:
            best_payoff = payoff
            j = -1

        return j

    def cross_over(self):
        self.children = []
        for n in range(0, self.population_size // 2):
            ind1 = random.choice(self.mating_pool)
            ind2 = random.choice(self.mating_pool)

            child = copy.deepcopy(ind1)
            child.id = self.id
            self.id += 1
            sub_solution = ind2.random_sub_solution()

            for x in sub_solution:
                child.remove_vertice(x)

            n = self.best_insertion(child, sub_solution)
            child.insert_sub_solution(sub_solution, n)
            child.cost = self.graph.get_cost(child.genes)
            self.children.append(child)

    def selection(self):
        self.mating_pool = []
        iterator_list = copy.deepcopy(self.individuals)
        while len(self.mating_pool) < self.population_size // 2:
            selected = min(iterator_list, key=lambda i: i.cost)
            iterator_list.remove(selected)
            self.mating_pool.append(selected)

    def mutation(self):
        for i in range(len(self.children)):
            prob = random.uniform(0, 1)
            if prob < self.mutation_probability:
                self.children[i].mutate()
                self.children[i].cost = self.graph.get_cost(self.children[i].genes)

    def substitution(self):
        new_population = []
        iterator_list = copy.deepcopy(self.individuals + self.children)

        while len(new_population) < self.population_size:
            selected = min(iterator_list, key=lambda i: i.cost)
            iterator_list.remove(selected)
            new_population.append(selected)

        self.individuals = new_population

    def evaluate(self):
        costs = [i.cost for i in self.individuals]
        mean_cost = sum(costs) / len(costs)
        min_cost = min(costs)
        if min_cost * 1.01 > mean_cost:
            print("BREAK")
            return True
        else:
            return False

    def run(self):
        start = datetime.datetime.utcnow()
        frac = 0.1
        plot = [10 * x for x in range(1, self.max_generations // 10 + 1)]
        initial_time = datetime.datetime.utcnow()
        for i in range(self.max_generations):
            if i / self.max_generations >= frac:
                print("Iteration ", str(i), " of ", str(self.max_generations))
                frac += 0.1
            # print("Generation: ", i + 1)
            self.selection()
            self.cross_over()
            self.mutation()
            self.substitution()
            self.update_best()

            if i >= self.max_generations // 5:
                if self.evaluate():
                    break

            if self.plot:
                if (i + 1) in plot:
                    filename = (
                        "plots/ga_"
                        + str(self.graph.number_vertices)
                        + "_"
                        + str(i + 1)
                        + ".png"
                    )
                    title = (
                        "Genetic algorithm. Iteration "
                        + str(i + 1)
                        + "\n Solution cost: "
                        + str(round(self.best_cost, 2))
                    )
                    self.graph.plot_solution(
                        self.best_genes,
                        pheromones=False,
                        filename=filename,
                        title=title,
                    )

            # self.get_best()
            if (datetime.datetime.utcnow() - initial_time).seconds > self.max_time:
                break

        self.time = datetime.datetime.utcnow() - start

        # print('FINISHED')

    def show(self):
        print("\nSOLUTION:")
        print(
            "Individuals: "
            + str(self.population_size)
            + ". Iterations: "
            + str(self.max_generations)
        )
        for ind in self.individuals:
            print("Best solution: %s\t|\tcost: %d" % (str(ind.genes), ind.cost))

    def save(self, file):
        best = self.individuals[0]
        for individual in self.individuals:
            if individual.cost < best.cost:
                best = individual
        text = (
            "GA. Best solution: "
            + str(best.genes)
            + "\t|\tCost: "
            + str(best.cost)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)

    def get_time(self):
        return self.time



from .aco import AntColonyOptimization
from .ga import GeneticAlgorithm
from .graph import Graph
from .lp import LinearIntegerProgram
from .pso import ParticleSwarmOptimization
from .sa import SimulatedAnnealing
from .two_opt import TwoOpt



import setuptools
import os

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    required = fh.read().splitlines()

extra_required = {"plot": list()}
with open("extra-requirements.txt", "r") as fh:
    extra_required["plot"] = fh.read().splitlines()


extra_required["plot"].remove("-r requirements.txt")

setuptools.setup(
    name="tsp-solvers",
    version="0.0.1",
    author="Guillermo González-Santander",
    author_email="g.gsantanderdelacruz@gmail.com",
    description="Set of different methods to solve the Travelling Salesman Problem. Each method has its own class",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/ggsdc/tsp-solvers",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=required,
    extras_require=extra_required,
)



import datetime
import math
import random
from itertools import accumulate


class Ant:
    def __init__(self, alpha, beta, number_vertices, graph):
        self.alpha = alpha
        self.beta = beta
        self.number_vertices = number_vertices
        self.graph = graph
        self.solution = None
        self.cost = 0
        self.unvisited_nodes = list(self.graph.vertices)

    # @profile
    def _select_node(self):

        pheromone = [
            math.pow(
                self.graph.edges[(self.solution[-1], unvisited)].pheromone, self.alpha
            )
            for unvisited in self.unvisited_nodes
        ]
        visibility = [
            math.pow(
                1 / self.graph.edges[(self.solution[-1], unvisited)].cost, self.beta
            )
            for unvisited in self.unvisited_nodes
        ]
        denominator = sum([x * y for x, y in zip(pheromone, visibility)])
        cumulative_probability = list(
            accumulate([p * h / denominator for p, h in zip(pheromone, visibility)])
        )
        random_value = random.uniform(0, 1)

        added_idx = next(
            idx
            for idx, value in enumerate(cumulative_probability)
            if value >= random_value
        )
        added = self.unvisited_nodes[added_idx]
        self.unvisited_nodes.remove(added)
        return added

    # @profile
    def find_solution(self):
        self.solution = [random.randint(0, self.number_vertices - 1)]
        self.unvisited_nodes = list(self.graph.vertices)
        self.unvisited_nodes.remove(self.solution[0])
        while len(self.solution) < self.number_vertices:
            self.solution.append(self._select_node())
        return self.solution

    def get_cost(self):
        self.cost = self.graph.get_cost(self.solution)
        return self.cost


class AntColonyOptimization:
    def __init__(
        self,
        mode="AS",
        graph=None,
        iterations=100,
        population_size=10,
        elitist_weight=1,
        min_scaling_factor=0.001,
        alpha=1,
        beta=3,
        rho=0.1,
        pheromone_deposit=1,
        initial_pheromone=1,
        max_time=60,
        plot=False,
    ):
        self.mode = mode
        self.graph = graph
        self.iterations = iterations
        self.population_size = population_size
        self.elitist_weight = elitist_weight
        self.min_scaling_factor = min_scaling_factor
        self.alpha = alpha
        self.beta = beta
        self.rho = rho
        self.pheromone_deposit = pheromone_deposit
        self.initial_pheromone = initial_pheromone
        self.number_vertices = self.graph.number_vertices
        self.max_time = max_time
        self.plot = plot
        self.time = None

        for edge in self.graph.edges:
            self.graph.update_pheromone(edge, initial_pheromone)

        self.ants = [
            Ant(alpha, beta, self.number_vertices, self.graph)
            for _ in range(self.population_size)
        ]

        self.best_solution = None
        self.best_cost = float("inf")

    # @profile
    def _add_pheromone(self, solution, cost, weight=1):
        pheromone_to_add = self.pheromone_deposit / cost
        for i in range(self.number_vertices):
            self.graph.edges[
                (solution[i], solution[(i + 1) % self.number_vertices])
            ].pheromone = (
                self.graph.edges[
                    (solution[i], solution[(i + 1) % self.number_vertices])
                ].pheromone
                + weight * pheromone_to_add
            )

    def get_best(self):
        best = self.ants[0]
        for ant in self.ants:
            if ant.cost < best.cost:
                best = ant
        print("\nACO:")
        print("Best solution: ", str(best.solution), "\t|\tcost: ", str(best.cost))

    # @profile
    def evaluate(self):
        costs = [i.cost for i in self.ants]
        mean_cost = sum(costs) / len(costs)
        min_cost = min(costs)

        if min_cost * 1.01 > mean_cost:
            print("BREAK")
            return True
        else:
            return False

    # @profile
    def _as(self):
        frac = 0.1
        plot = [10 * x for x in range(1, self.iterations // 10 + 1)]
        initial_time = datetime.datetime.utcnow()
        for i in range(self.iterations):
            if i / self.iterations >= frac:
                print("Iteration ", str(i), " of ", str(self.iterations))
                frac += 0.1

            if i != 0:
                for edge in self.graph.edges:
                    self.graph.edges[edge].pheromone = self.graph.edges[
                        edge
                    ].pheromone * (1 - self.rho)

            for ant in self.ants:
                self._add_pheromone(ant.find_solution(), ant.get_cost())
                if ant.cost < self.best_cost:
                    self.best_solution = ant.solution
                    self.best_cost = ant.cost

            if i >= self.iterations // 5:
                if self.evaluate():
                    break

            if self.plot:
                if (i + 1) in plot:
                    filename = (
                        "plots/aco_"
                        + str(self.graph.number_vertices)
                        + "_"
                        + str(i + 1)
                        + ".png"
                    )
                    title = (
                        "Ant Colony Optimization. Iteration "
                        + str(i + 1)
                        + "\n Solution cost: "
                        + str(round(self.best_cost, 2))
                    )
                    self.graph.plot_solution(
                        self.best_solution,
                        pheromones=True,
                        filename=filename,
                        title=title,
                    )

            if (datetime.datetime.utcnow() - initial_time).seconds > self.max_time:
                break

    def run(self):
        start = datetime.datetime.utcnow()
        if self.mode == "AS":
            self._as()
        else:
            pass
        self.time = datetime.datetime.utcnow() - start

    def show(self):
        print("\nSOLUTION:")
        print(
            "Ants: "
            + str(self.population_size)
            + ". Iterations: "
            + str(self.iterations)
            + "\n"
        )
        for ant in self.ants:
            print("Best solution: %s\t|\tcost: %d" % (str(ant.solution), ant.cost))

    def save(self, file):
        best = self.ants[0]
        for ant in self.ants:
            if ant.cost < best.cost:
                best = ant
        text = (
            "ACO. Best solution: "
            + str(best.solution)
            + "\t|\tCost: "
            + str(best.cost)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)

    def get_time(self):
        return self.time
